import os

import geopandas as gpd

from waterpyk.raster import add_slash


def reproject(save_dir, shapefile_path, target_epsg='4326'):
    """"
    Reproject a shapefile using geopandas. Save reprojected shapefile with ending '_[target_epsg]'. 
    Returns path to reprojected shapefile.

    Args:
        save_dir (str): path to directory to save reprojected shapefile.
        shapefile_path (str): path to original file.
        target_epsg (str, optional): EPSG code to reproject to. Defaults to '4326'.

    Returns:
        str: path to reprojected shapefile
    """
    # Perform checks
    if type(target_epsg) != str:
        raise TypeError(f'target_epsg must be str. Got {type(target_epsg)}')
    if os.path.exists(shapefile_path) == False:
        raise FileNotFoundError(f'{shapefile_path} does not exist.')
    if os.path.exists(save_dir) == False:
        raise FileNotFoundError(f'{save_dir} does not exist.')

    # Read shapefile, get name, and check if path has '/' at end
    gdf = gpd.read_file(shapefile_path)
    file_name = str(shapefile_path.split('/')[-1]).split('.')[0]
    save_dir = add_slash(save_dir)

    # Reproject geodataframe if necessary
    if gdf.crs != 'epsg:' + target_epsg:
        print(f'Reprojecting {file_name}.shp from {gdf.crs} to {target_epsg}')
        gdf = gdf.to_crs(epsg=target_epsg)
        save_path = save_dir + file_name + '_epsg' + target_epsg + '.shp'
        gdf.to_file(save_path)

    else:
        print(f'No reproject: {file_name}.shp already in {target_epsg}).')
        save_path = save_dir

    return save_path



from waterpyk import main


def test_point():
    point = main.StudyArea([1111, 1111], 'minimal')
    assert point.kind == 'point', "Lat/Long not being categorized as 'point'"
    assert point.site_name == ''
    assert point.description == ''
    
    



# Base class: Inheriting all other errors from this makes it easier to turn all of them
# off if someone wishes to later down the road, without interfering with other modules' error messages.
class BaseValidateError(ValueError):
    pass

# Specific errors used in module
class GageTooLongError(BaseValidateError):
    pass

class NoDateSpecifiedError(BaseValidateError):
    pass

class MissingBandsError(BaseValidateError):
    pass






import os
from glob import glob

import fiona
import geopandas as gpd
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import rasterio as rio
from rasterio.mask import mask

from waterpyk import raster
from waterpyk import shapefile as shp


def file_finder(base_directory, file_type, skip_files=None, include_files=None, sub_dirs=4):
    """
    Get all of the files with designated file types (and filtered by include/skip files lists)
    for all of the subdirectories in the base directory. Get a list of paths and the names of those files.

    Args:
        base_directory (str): Directory to search for files in. Don't include the / at the end.
        file_type (str or list of str): File type(s) to search for. Include the period, ie '.tif' or provide a list of strings, like ['.tif', '.tiff'].
        skip_files (list of str, optional): List of files to skip. Just include the name of the file and the ending, i.e. ['image_of_cats.tif'] and not any folder/location information.
        include_files (list of str, optional): List of files to keep in the final list. Just include the name of the file and the ending, i.e. ['image_of_cats.tif'] and not any folder/location information.
        sub_dirs (int, optional): Number of sub-directory levels to look for files in. Default = 4.

    Returns:
        list, list: list of file locations and list of file names

    """
    def glob_options(file_type):
        endings = []
        for i in range(sub_dirs):
            endings.append('/**'*i + '/*' + file_type)
        return endings

    # If there is one file type, just get the glob_options
    if type(file_type) == str:
        endings = glob_options(file_type)
    # If there is a list of file types, get glob_options for all items in the list
    elif type(file_type) == list:
        endings = []
        for item in file_type:
            if type(item) == str:
                endings.append(glob_options(item))
            else:
                raise TypeError(
                    f'File_type can take a str (ex ".tif") or a list of strings. Recieved list of {type(item)}')
        # Flatten endings list
        endings = [item for sublist in endings for item in sublist]
    # If it is not a list of strings or a string, raise error
    else:
        raise TypeError(
            f'File_type can take a str (ex ".tif") or a list of strings. Recieved {type(file_type)}')

    file_paths = []
    for ending in endings:
        file_paths.append(glob(base_directory + ending))

    # Flatten file_paths list
    file_paths = [item for sublist in file_paths for item in sublist]

    # Remove skip_files
    if skip_files is not None:
        file_paths = [i for i in file_paths if i.split(
            '/')[-1] not in skip_files]

    # Remove all files not in include_files
    if include_files is not None:
        file_paths = [i for i in file_paths if i.split(
            '/')[-1] in include_files]

    # Get a name for each file for plotting, etc
    file_names = [str(i.split('/')[-1]).split('.')[0] for i in file_paths]

    return file_paths, file_names


def clip_raster_by_shapefile(raster_path, shapefile_path, target_epsg='4326', raster_save_dir='.', shape_save_dir='.', fig_save_dir='.', plot_result=False):
    # Reproject shapefile (will skip if not necessary
    shapefile_path = shp.reproject(save_dir=shape_save_dir,
                                   shapefile_path=shapefile_path,
                                   target_epsg=target_epsg)
    # Get geometry of shapefile
    with fiona.open(shapefile_path, "r") as shapefile:
        shapes = [feature["geometry"] for feature in shapefile]

    # Reproject raster (will skip if not necessary)
    raster_path = raster.reproject(save_dir=raster_save_dir,
                                   raster_path=raster_path,
                                   target_epsg=target_epsg)

    # Open reprojected raster and clip by reprojected hapefile
    with rio.open(raster_path) as src:
        out_image, out_transform = mask(dataset=src,
                                        shapes=shapes,
                                        crop=True)
        out_meta = src.meta

    # Update the metadata of the clipped raster
    out_meta.update({"driver": "GTiff",
                     "height": out_image.shape[1],
                     "width": out_image.shape[2],
                     "transform": out_transform})

    # Get a new name for the final raster
    name_r = str(raster_path.split('/')[-1]).split('.')[0]
    name_s = str(shapefile_path.split('/')[-1]).split('.')[0]
    dir_r = raster.add_slash(raster_save_dir)
    raster_save_path = dir_r + name_r + '_maskedby_' + name_s + '.tif'

    # Save the clipped raster to a new file
    with rio.open(raster_save_path, "w", **out_meta) as dest:
        dest.write(out_image)

    if plot_result == True:
        with rio.open(raster_save_path) as src:
            plt.imshow(src.read(1))
            dir_fig = raster.add_slash(fig_save_dir)
            plt.savefig(dir_fig + name_r + '_' + name_s + '.png')
